#!/usr/bin/env python3
"""
Intrusion Detection System
Advanced network intrusion detection and monitoring tool.
Parameters:
    - interface: Network interface to monitor
    - mode: Detection mode (passive, active)
    - rules: Rules configuration file
Requires: scapy, psutil, netfilterqueue packages
"""
import argparse
import logging
from datetime import datetime
import json
import os
import sys
from pathlib import Path
import socket
import struct
import threading
import queue
import time
from collections import defaultdict
import ipaddress
from scapy.all import *
from scapy.layers.http import *
from scapy.layers.inet import *
import psutil
from netfilterqueue import NetfilterQueue

class IntrusionDetection:
    def __init__(self, interface, mode='passive', rules_file=None):
        self.interface = interface
        self.mode = mode.lower()
        self.rules_file = rules_file
        self.results = {}
        self.packet_queue = queue.Queue()
        self.stop_flag = threading.Event()
        self.setup_logging()
        self.load_rules()
        self.initialize_stats()

    def setup_logging(self):
        logging.basicConfig(
            level=logging.INFO,
            format='%(asctime)s - %(levelname)s - %(message)s'
        )
        self.logger = logging.getLogger(__name__)

    def load_rules(self):
        """Load detection rules from configuration"""
        if self.rules_file and os.path.exists(self.rules_file):
            try:
                with open(self.rules_file, 'r') as f:
                    self.rules = json.load(f)
            except Exception as e:
                self.logger.error(
                    f"Error loading rules file: {str(e)}"
                )
                self.rules = self.get_default_rules()
        else:
            self.rules = self.get_default_rules()

    def get_default_rules(self):
        """Get default detection rules"""
        return {
            'ip_blacklist': [
                '192.168.1.100',
                '10.0.0.100'
            ],
            'port_blacklist': [
                22, 23, 3389
            ],
            'protocol_blacklist': [
                'telnet'
            ],
            'packet_thresholds': {
                'syn_flood': 100,
                'icmp_flood': 50,
                'port_scan': 20
            },
            'content_patterns': [
                r'(?i)admin',
                r'(?i)password',
                r'(?i)select.*from',
                r'(?i)union.*select'
            ],
            'time_windows': {
                'syn_flood': 5,
                'icmp_flood': 5,
                'port_scan': 10
            }
        }

    def initialize_stats(self):
        """Initialize statistics counters"""
        self.stats = {
            'total_packets': 0,
            'alerts': defaultdict(int),
            'protocols': defaultdict(int),
            'src_ips': defaultdict(int),
            'dst_ips': defaultdict(int),
            'src_ports': defaultdict(int),
            'dst_ports': defaultdict(int),
            'packet_sizes': [],
            'start_time': datetime.now(),
            'syn_flood_counter': defaultdict(int),
            'icmp_flood_counter': defaultdict(int),
            'port_scan_counter': defaultdict(lambda: defaultdict(int))
        }

    def packet_callback(self, packet):
        """Process captured packet"""
        try:
            # Update basic statistics
            self.stats['total_packets'] += 1
            
            # Extract IP layer
            if IP in packet:
                ip_src = packet[IP].src
                ip_dst = packet[IP].dst
                self.stats['src_ips'][ip_src] += 1
                self.stats['dst_ips'][ip_dst] += 1
                
                # Check IP blacklist
                if (ip_src in self.rules['ip_blacklist'] or
                    ip_dst in self.rules['ip_blacklist']):
                    self.generate_alert(
                        'blacklisted_ip',
                        f"Blacklisted IP detected: {ip_src} -> {ip_dst}"
                    )

                # Extract transport layer
                if TCP in packet:
                    self.process_tcp_packet(packet)
                elif UDP in packet:
                    self.process_udp_packet(packet)
                elif ICMP in packet:
                    self.process_icmp_packet(packet)

                # Update packet size statistics
                self.stats['packet_sizes'].append(len(packet))

        except Exception as e:
            self.logger.error(f"Error processing packet: {str(e)}")

    def process_tcp_packet(self, packet):
        """Process TCP packet"""
        try:
            ip_src = packet[IP].src
            ip_dst = packet[IP].dst
            sport = packet[TCP].sport
            dport = packet[TCP].dport
            
            self.stats['protocols']['TCP'] += 1
            self.stats['src_ports'][sport] += 1
            self.stats['dst_ports'][dport] += 1

            # Check for SYN flood
            if packet[TCP].flags & 0x02:  # SYN flag
                self.stats['syn_flood_counter'][ip_src] += 1
                if self.detect_syn_flood(ip_src):
                    self.generate_alert(
                        'syn_flood',
                        f"Possible SYN flood from {ip_src}"
                    )

            # Check for port scan
            self.stats['port_scan_counter'][ip_src][dport] += 1
            if self.detect_port_scan(ip_src):
                self.generate_alert(
                    'port_scan',
                    f"Possible port scan from {ip_src}"
                )

            # Check port blacklist
            if (sport in self.rules['port_blacklist'] or
                dport in self.rules['port_blacklist']):
                self.generate_alert(
                    'blacklisted_port',
                    f"Blacklisted port accessed: {sport} -> {dport}"
                )

            # Deep packet inspection
            if Raw in packet:
                payload = packet[Raw].load
                self.inspect_payload(payload, ip_src, ip_dst)

        except Exception as e:
            self.logger.error(f"Error processing TCP packet: {str(e)}")

    def process_udp_packet(self, packet):
        """Process UDP packet"""
        try:
            ip_src = packet[IP].src
            ip_dst = packet[IP].dst
            sport = packet[UDP].sport
            dport = packet[UDP].dport
            
            self.stats['protocols']['UDP'] += 1
            self.stats['src_ports'][sport] += 1
            self.stats['dst_ports'][dport] += 1

            # Check port blacklist
            if (sport in self.rules['port_blacklist'] or
                dport in self.rules['port_blacklist']):
                self.generate_alert(
                    'blacklisted_port',
                    f"Blacklisted port accessed: {sport} -> {dport}"
                )

            # Deep packet inspection
            if Raw in packet:
                payload = packet[Raw].load
                self.inspect_payload(payload, ip_src, ip_dst)

        except Exception as e:
            self.logger.error(f"Error processing UDP packet: {str(e)}")

    def process_icmp_packet(self, packet):
        """Process ICMP packet"""
        try:
            ip_src = packet[IP].src
            ip_dst = packet[IP].dst
            
            self.stats['protocols']['ICMP'] += 1
            self.stats['icmp_flood_counter'][ip_src] += 1

            # Check for ICMP flood
            if self.detect_icmp_flood(ip_src):
                self.generate_alert(
                    'icmp_flood',
                    f"Possible ICMP flood from {ip_src}"
                )

        except Exception as e:
            self.logger.error(f"Error processing ICMP packet: {str(e)}")

    def inspect_payload(self, payload, ip_src, ip_dst):
        """Inspect packet payload for suspicious content"""
        try:
            payload_str = payload.decode('utf-8', errors='ignore')
            
            for pattern in self.rules['content_patterns']:
                if re.search(pattern, payload_str):
                    self.generate_alert(
                        'suspicious_content',
                        f"Suspicious content detected: {ip_src} -> {ip_dst}"
                    )
                    break

        except Exception as e:
            self.logger.error(f"Error inspecting payload: {str(e)}")

    def detect_syn_flood(self, ip_src):
        """Detect SYN flood attack"""
        threshold = self.rules['packet_thresholds']['syn_flood']
        window = self.rules['time_windows']['syn_flood']
        
        return (
            self.stats['syn_flood_counter'][ip_src] > threshold
            and (datetime.now() - self.stats['start_time']).seconds <= window
        )

    def detect_icmp_flood(self, ip_src):
        """Detect ICMP flood attack"""
        threshold = self.rules['packet_thresholds']['icmp_flood']
        window = self.rules['time_windows']['icmp_flood']
        
        return (
            self.stats['icmp_flood_counter'][ip_src] > threshold
            and (datetime.now() - self.stats['start_time']).seconds <= window
        )

    def detect_port_scan(self, ip_src):
        """Detect port scanning activity"""
        threshold = self.rules['packet_thresholds']['port_scan']
        window = self.rules['time_windows']['port_scan']
        
        unique_ports = len(self.stats['port_scan_counter'][ip_src])
        return (
            unique_ports > threshold
            and (datetime.now() - self.stats['start_time']).seconds <= window
        )

    def generate_alert(self, alert_type, message):
        """Generate security alert"""
        self.stats['alerts'][alert_type] += 1
        self.logger.warning(f"ALERT: {message}")
        
        # Log alert to file
        try:
            with open('ids_alerts.log', 'a') as f:
                f.write(f"{datetime.now().isoformat()} - {alert_type}: {message}\n")
        except Exception as e:
            self.logger.error(f"Error logging alert: {str(e)}")

    def packet_handler(self):
        """Handle captured packets"""
        try:
            while not self.stop_flag.is_set():
                try:
                    packet = self.packet_queue.get(timeout=1)
                    self.packet_callback(packet)
                except queue.Empty:
                    continue
        except Exception as e:
            self.logger.error(f"Error in packet handler: {str(e)}")

    def start_capture(self):
        """Start packet capture"""
        try:
            # Start packet handler thread
            handler_thread = threading.Thread(
                target=self.packet_handler
            )
            handler_thread.start()

            # Start packet capture
            sniff(
                iface=self.interface,
                prn=lambda x: self.packet_queue.put(x),
                store=0,
                stop_filter=lambda x: self.stop_flag.is_set()
            )

            # Wait for handler to finish
            handler_thread.join()

        except Exception as e:
            self.logger.error(f"Error starting capture: {str(e)}")

    def generate_report(self):
        """Generate intrusion detection report"""
        try:
            report = {
                'timestamp': datetime.now().isoformat(),
                'interface': self.interface,
                'mode': self.mode,
                'duration': str(
                    datetime.now() - self.stats['start_time']
                ),
                'statistics': {
                    'total_packets': self.stats['total_packets'],
                    'alerts': dict(self.stats['alerts']),
                    'protocols': dict(self.stats['protocols']),
                    'top_source_ips': dict(
                        sorted(
                            self.stats['src_ips'].items(),
                            key=lambda x: x[1],
                            reverse=True
                        )[:10]
                    ),
                    'top_destination_ips': dict(
                        sorted(
                            self.stats['dst_ips'].items(),
                            key=lambda x: x[1],
                            reverse=True
                        )[:10]
                    ),
                    'top_source_ports': dict(
                        sorted(
                            self.stats['src_ports'].items(),
                            key=lambda x: x[1],
                            reverse=True
                        )[:10]
                    ),
                    'top_destination_ports': dict(
                        sorted(
                            self.stats['dst_ports'].items(),
                            key=lambda x: x[1],
                            reverse=True
                        )[:10]
                    ),
                    'average_packet_size': (
                        sum(self.stats['packet_sizes']) /
                        len(self.stats['packet_sizes'])
                        if self.stats['packet_sizes'] else 0
                    )
                }
            }
            return report
        except Exception as e:
            self.logger.error(f"Error generating report: {str(e)}")
            return None

    def save_report(self, report, output_file):
        """Save detection report to file"""
        try:
            with open(output_file, 'w') as f:
                json.dump(report, f, indent=4)
            self.logger.info(f"Report saved to {output_file}")
        except Exception as e:
            self.logger.error(f"Error saving report: {str(e)}")

    def print_summary(self, report):
        """Print detection summary"""
        self.logger.info("\nIntrusion Detection Summary:")
        self.logger.info(f"Interface: {report['interface']}")
        self.logger.info(f"Mode: {report['mode']}")
        self.logger.info(f"Duration: {report['duration']}")
        
        if 'statistics' in report:
            stats = report['statistics']
            self.logger.info("\nTraffic Statistics:")
            self.logger.info(
                f"Total Packets: {stats['total_packets']:,}"
            )
            self.logger.info(
                f"Average Packet Size: {stats['average_packet_size']:.2f} bytes"
            )
            
            self.logger.info("\nProtocol Distribution:")
            for protocol, count in stats['protocols'].items():
                self.logger.info(f"{protocol}: {count:,}")
            
            if stats['alerts']:
                self.logger.info("\nAlerts Generated:")
                for alert_type, count in stats['alerts'].items():
                    self.logger.info(
                        f"{alert_type}: {count:,}"
                    )
            
            self.logger.info("\nTop Source IPs:")
            for ip, count in stats['top_source_ips'].items():
                self.logger.info(f"{ip}: {count:,}")
            
            self.logger.info("\nTop Destination IPs:")
            for ip, count in stats['top_destination_ips'].items():
                self.logger.info(f"{ip}: {count:,}")

    def start_detection(self, output_file=None):
        """Start intrusion detection"""
        self.logger.info("Starting intrusion detection")
        self.logger.info(f"Mode: {self.mode}")
        self.logger.info(f"Interface: {self.interface}")
        
        try:
            # Start packet capture
            self.start_capture()

            # Generate and save report
            report = self.generate_report()
            if report and output_file:
                self.save_report(report, output_file)
            if report:
                self.print_summary(report)
                
        except KeyboardInterrupt:
            self.logger.info("Detection interrupted by user")
            self.stop_flag.set()
        except Exception as e:
            self.logger.error(f"Error during detection: {str(e)}")
            self.stop_flag.set()

if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Intrusion Detection System"
    )
    parser.add_argument(
        "interface",
        help="Network interface to monitor"
    )
    parser.add_argument(
        "--mode",
        choices=['passive', 'active'],
        default='passive',
        help="Detection mode"
    )
    parser.add_argument(
        "--rules",
        help="Rules configuration file"
    )
    parser.add_argument(
        "--output",
        help="Output file for detection report"
    )
    args = parser.parse_args()

    ids = IntrusionDetection(
        args.interface,
        args.mode,
        args.rules
    )
    ids.start_detection(args.output)
